@@ -9,12 +9,21 @@ import uuid
 import time
 import sys
 import os
+import functools
 
 from . import bridge
 
 if sys.version_info[0] == 2:
     from socket import error as ConnectionRefusedError  # ConnectionRefusedError not defined in python2, this is next closest thing
 
+def print_stats(func):
+    @functools.wraps(func)
+    def wrapper(self, *args, **kwargs):
+        start_stats = self.test_bridge.get_stats()
+        func(self, *args, **kwargs)
+        print("\n{}:\n\t{}\n".format(func.__name__, self.test_bridge.get_stats()-start_stats))
+
+    return wrapper
 
 class TestBridge(unittest.TestCase):
     """ Assumes there's a bridge server running at DEFAULT_SERVER_PORT """
@@ -22,26 +31,30 @@ class TestBridge(unittest.TestCase):
     @classmethod
     def setUpClass(cls):
         port = int(os.environ.get("TEST_PORT", bridge.DEFAULT_SERVER_PORT))
-        TestBridge.test_bridge = bridge.BridgeClient(
-            connect_to_port=port, loglevel=logging.DEBUG)
+        cls.test_bridge = bridge.BridgeClient(
+            connect_to_port=port, loglevel=logging.DEBUG, record_stats=True)
 
+    @print_stats
     def test_import(self):
-
-        mod = TestBridge.test_bridge.remote_import("base64")
+        start_stats = self.test_bridge.get_stats()
+        mod = self.test_bridge.remote_import("base64")
         self.assertTrue(mod is not None)
+        total_stats = self.test_bridge.get_stats() - start_stats
 
+    @print_stats
     def test_call_no_args(self):
 
-        mod = TestBridge.test_bridge.remote_import("uuid")
+        mod = self.test_bridge.remote_import("uuid")
 
         result = mod.uuid4()
 
         self.assertTrue(result is not None)
 
+    @print_stats
     def test_call_arg(self):
         # also tests call with bytestring arg in python3
 
-        mod = TestBridge.test_bridge.remote_import("base64")
+        mod = self.test_bridge.remote_import("base64")
 
         test_str = str(uuid.uuid4())
 
@@ -52,14 +65,15 @@ class TestBridge(unittest.TestCase):
         else:
             # python2 can't send a byte string, and if the other end is python3, b64encode won't work on a string.
             # instead we'll try creating a uuid from the string
-            remote_uuid = TestBridge.test_bridge.remote_import("uuid")
+            remote_uuid = self.test_bridge.remote_import("uuid")
             new_uuid = remote_uuid.UUID(test_str)
             result_str = str(new_uuid)
 
         self.assertEqual(test_str, result_str)
 
+    @print_stats
     def test_call_multi_args(self):
-        mod = TestBridge.test_bridge.remote_import("re")
+        mod = self.test_bridge.remote_import("re")
 
         remote_obj = mod.compile("foo", mod.IGNORECASE)
 
@@ -67,19 +81,21 @@ class TestBridge(unittest.TestCase):
 
         self.assertTrue(remote_obj.match("FOO") is not None)
 
+    @print_stats
     def test_call_with_remote_obj(self):
 
-        mod = TestBridge.test_bridge.remote_import("uuid")
+        mod = self.test_bridge.remote_import("uuid")
 
         remote_obj = mod.uuid4()
         result = str(remote_obj)
         self.assertTrue(result is not None)
         self.assertTrue("-" in result and "4" in result)
 
+    @print_stats
     def test_call_with_str(self):
         """ also tests calling str() on remote obj """
 
-        mod = TestBridge.test_bridge.remote_import("uuid")
+        mod = self.test_bridge.remote_import("uuid")
 
         test_uuid_str = "00010203-0405-0607-0809-0a0b0c0d0e0f"
 
@@ -92,40 +108,46 @@ class TestBridge(unittest.TestCase):
     # set a function into the remote __main__/globals() to call
     # callback as key func in list.sort
 
+    @print_stats
     def test_call_kwargs(self):
         self.skipTest("Not implemented yet")
 
+    @print_stats
     def test_get(self):
-        mod = TestBridge.test_bridge.remote_import("uuid")
+        mod = self.test_bridge.remote_import("uuid")
         remote_doc = mod.__doc__
         self.assertTrue("RFC 4122" in remote_doc)
 
+    @print_stats
     def test_set(self):
         test_string = "hello world"
-        mod = TestBridge.test_bridge.remote_import("__main__")
+        mod = self.test_bridge.remote_import("__main__")
         mod.test = test_string
 
         self.assertEqual(test_string, mod.test)
 
+    @print_stats
     def test_get_non_existent(self):
         """ Check that requesting a non-existent attribute over the bridge raises an attributeerror """
-        mod = TestBridge.test_bridge.remote_import("re")
+        mod = self.test_bridge.remote_import("re")
 
         remote_obj = mod.compile("foo")
 
         with self.assertRaises(AttributeError):
             remote_obj.doesnt_exist
 
+    @print_stats
     def test_get_callable(self):
-        mod = TestBridge.test_bridge.remote_import("re")
+        mod = self.test_bridge.remote_import("re")
 
         remote_obj = mod.compile("foo")
 
         remote_callable = remote_obj.search
         self.assertTrue(isinstance(remote_callable, bridge.BridgedCallable))
 
+    @print_stats
     def test_callable(self):
-        mod = TestBridge.test_bridge.remote_import("re")
+        mod = self.test_bridge.remote_import("re")
 
         remote_obj = mod.compile("foo")
 
@@ -133,8 +155,9 @@ class TestBridge(unittest.TestCase):
 
         self.assertTrue(remote_callable("fooa") is not None)
 
+    @print_stats
     def test_serialize_deserialize_types(self):
-        mod = TestBridge.test_bridge.remote_import("__main__")
+        mod = self.test_bridge.remote_import("__main__")
         remote_list = mod.__builtins__.list
 
         # assemble a list of different types
@@ -153,9 +176,10 @@ class TestBridge(unittest.TestCase):
         # check it's the same
         self.assertEqual(test_list, created_list)
 
+    @print_stats
     def test_serialize_deserialize_bytes(self):
         """ byte strings across 2<->3 bridges will be forced to strings (because py2 treats bytes and strs as the same thing """
-        mod = TestBridge.test_bridge.remote_import("__main__")
+        mod = self.test_bridge.remote_import("__main__")
         remote_list = mod.__builtins__.list
 
         test_list = [b"bytes"]
@@ -167,12 +191,14 @@ class TestBridge(unittest.TestCase):
         self.assertTrue(created_list[0] == test_list[0]
                         or created_list[0] == test_list[0].decode("utf-8"))
 
+    @print_stats
     def test_serialize_deserialize_bridge_object(self):
         # bridge objects TODO
         self.skipTest("Not implemented yet")
 
+    @print_stats
     def test_none_result(self):
-        mod = TestBridge.test_bridge.remote_import("re")
+        mod = self.test_bridge.remote_import("re")
 
         remote_obj = mod.compile("foo")
 
@@ -180,15 +206,17 @@ class TestBridge(unittest.TestCase):
 
         self.assertTrue(remote_callable("abar") is None)
 
+    @print_stats
     def test_exception(self):
         self.skipTest("Not implemented yet")
 
+    @print_stats
     def test_callback(self):
         """ Test we correctly handle calling back to here from across the bridge """
         def sort_fn(val):
             return len(val)
 
-        mod = TestBridge.test_bridge.remote_import("__main__")
+        mod = self.test_bridge.remote_import("__main__")
         remote_sorted = mod.__builtins__.sorted
 
         test_list = ["aaa", "bb", "c"]
@@ -196,9 +224,10 @@ class TestBridge(unittest.TestCase):
 
         self.assertEqual(sorted(test_list, key=sort_fn), sorted_list)
 
+    @print_stats
     def test_remote_iterable(self):
         """ Test we can access values from a remote iterable """
-        mod = TestBridge.test_bridge.remote_import("__main__")
+        mod = self.test_bridge.remote_import("__main__")
         remote_range = mod.__builtins__.range
 
         remote_it = remote_range(4, 10, 2)
@@ -207,9 +236,10 @@ class TestBridge(unittest.TestCase):
 
         self.assertEqual(list(range(4, 10, 2)), it_values)
 
+    @print_stats
     def test_remote_iterable_for(self):
         """ Test we can access values from a remote iterable with a for loop """
-        mod = TestBridge.test_bridge.remote_import("__main__")
+        mod = self.test_bridge.remote_import("__main__")
         remote_range = mod.__builtins__.range
 
         remote_it = remote_range(4, 10, 2)
@@ -219,13 +249,15 @@ class TestBridge(unittest.TestCase):
 
         self.assertEqual(list(range(4, 10, 2)), it_values)
 
+    @print_stats
     def test_float(self):
         """ Test we can sent a float value """
-        remote_time = TestBridge.test_bridge.remote_import("time")
+        remote_time = self.test_bridge.remote_import("time")
         remote_time.sleep(0.1)
 
+    @print_stats
     def test_is_bridged_object(self):
-        remote_uuid = TestBridge.test_bridge.remote_import("uuid")
+        remote_uuid = self.test_bridge.remote_import("uuid")
 
         remote_obj = remote_uuid.uuid4()
         local_obj = uuid.uuid4()
@@ -233,11 +265,12 @@ class TestBridge(unittest.TestCase):
         self.assertTrue(bridge._is_bridged_object(remote_obj))
         self.assertFalse(bridge._is_bridged_object(local_obj))
 
+    @print_stats
     def test_bridged_isinstance(self):
-        mod = TestBridge.test_bridge.remote_import("__main__")
+        mod = self.test_bridge.remote_import("__main__")
         remote_float = mod.__builtins__.float
         remote_int = mod.__builtins__.int
-        remote_uuid = TestBridge.test_bridge.remote_import("uuid")
+        remote_uuid = self.test_bridge.remote_import("uuid")
         remote_class = remote_uuid.UUID
         remote_obj = remote_uuid.uuid4()
         local_class = uuid.UUID
@@ -288,37 +321,43 @@ class TestBridge(unittest.TestCase):
         self.assertFalse(bridge.bridged_isinstance(
             remote_obj, (remote_float, remote_int)))
 
+    @print_stats
     def test_bridged_get_type(self):
         """ Make sure we can get an object representing the type of a bridged object """
-        remote_uuid = TestBridge.test_bridge.remote_import("uuid")
+        remote_uuid = self.test_bridge.remote_import("uuid")
         remote_obj = remote_uuid.uuid4()
 
         self.assertTrue("<class 'uuid.UUID'>" in str(remote_obj._bridged_get_type()))
         self.assertTrue("'type'" in str(remote_obj._bridged_get_type()._bridged_get_type()))
 
+    @print_stats
     def test_remote_eval(self):
-        self.assertEquals(3, TestBridge.test_bridge.remote_eval("1+2"))
+        self.assertEquals(3, self.test_bridge.remote_eval("1+2"))
 
+    @print_stats
     def test_remote_eval_bad_code(self):
         with self.assertRaises(bridge.BridgeException):
-            TestBridge.test_bridge.remote_eval("1+x")
+            self.test_bridge.remote_eval("1+x")
 
+    @print_stats
     def test_remote_eval_kwargs(self):
-        self.assertEquals(3, TestBridge.test_bridge.remote_eval("x+y", x=1, y=2))
+        self.assertEquals(3, self.test_bridge.remote_eval("x+y", x=1, y=2))
 
+    @print_stats
     def test_remote_eval_timeout(self):
-        remote_time = TestBridge.test_bridge.remote_import("time")
+        remote_time = self.test_bridge.remote_import("time")
 
         # check that it times out if not enough time allocated
         with self.assertRaises(Exception):
-            TestBridge.test_bridge.remote_eval("sleep(2)", timeout_override=1, sleep=remote_time.sleep)
+            self.test_bridge.remote_eval("sleep(2)", timeout_override=1, sleep=remote_time.sleep)
 
         # check that it works with enough time
-        TestBridge.test_bridge.remote_eval("sleep(2)", timeout_override=3, sleep=remote_time.sleep)
+        self.test_bridge.remote_eval("sleep(2)", timeout_override=3, sleep=remote_time.sleep)
 
+    @print_stats
     def test_operators(self):
         # check we can handle operator comparisons, addition, etc
-        remote_datetime = TestBridge.test_bridge.remote_import("datetime")
+        remote_datetime = self.test_bridge.remote_import("datetime")
         td1 = remote_datetime.timedelta(1)
         td2 = remote_datetime.timedelta(2)
 
@@ -327,26 +366,29 @@ class TestBridge(unittest.TestCase):
         self.assertEquals(remote_datetime.timedelta(3), td1 + td2)
         self.assertEquals(td1, td2//2)  # we use floordiv here, truediv tested below
 
+    @print_stats
     def test_truediv(self):
         # check we cleanly fallback from truediv to div
         # timedelta in jython2.7 implements __div__ but not __truediv__
-        remote_datetime = TestBridge.test_bridge.remote_import("datetime")
+        remote_datetime = self.test_bridge.remote_import("datetime")
         td1 = remote_datetime.timedelta(1)
         td2 = remote_datetime.timedelta(2)
         self.assertEquals(td1, td2/2)
 
+    @print_stats
     def test_len(self):
         # check we can handle len
-        remote_collections = TestBridge.test_bridge.remote_import("collections")
+        remote_collections = self.test_bridge.remote_import("collections")
         dq = remote_collections.deque()
         dq.append(1)
         dq.append(2)
         dq.append(3)
         self.assertEquals(3, len(dq))
 
+    @print_stats
     def test_bool(self):
         """ check we handle truthiness """
-        remote_collections = TestBridge.test_bridge.remote_import("collections")
+        remote_collections = self.test_bridge.remote_import("collections")
         dq = remote_collections.deque()
         self.assertFalse(bool(dq))
 
@@ -363,13 +405,14 @@ class TestBridge(unittest.TestCase):
             __nonzero__ = __bool__
 
         f = x(3)
-        self.assertFalse(TestBridge.test_bridge.remote_eval("bool(f)", f=f))
+        self.assertFalse(self.test_bridge.remote_eval("bool(f)", f=f))
         t = x(2)
-        self.assertTrue(TestBridge.test_bridge.remote_eval("bool(t)", t=t))
+        self.assertTrue(self.test_bridge.remote_eval("bool(t)", t=t))
 
+    @print_stats
     def test_bytes(self):
         """ Test that we handle calling bytes() on a bridged object """
-        remote_collections = TestBridge.test_bridge.remote_import("collections")
+        remote_collections = self.test_bridge.remote_import("collections")
         dq = remote_collections.deque()
         dq.append(1)
 
@@ -378,6 +421,31 @@ class TestBridge(unittest.TestCase):
             self.assertEquals(bytes(dq), "deque([1])")
         else:
             self.assertEquals(bytes(dq), b"\x01")
+            
+    @print_stats
+    def test_remote_inheritance(self):
+        """ check that we can inherit from a remote type """
+        remote_collections = self.test_bridge.remote_import("collections")
+        remote_deque = remote_collections.deque
+        
+        class new_deque(remote_deque):
+            def __init__(self, test):
+                remote_deque.__init__(self)
+                self.test = test
+                self.called = False
+                
+            def append(self, x):
+                remote_deque.append(self, x)
+                self.called = True
+                
+        nd = new_deque("test")
+        self.assertEquals(nd.test, "test")
+        
+        nd.append(1)
+        self.assertTrue(nd.called)
+        self.assertEquals(nd.pop(), 1)
+        
+        self.assertTrue(not isinstance(nd.append, bridge.BridgedCallable), "Expected local implementation to stay local - is actually: " + str(type(nd.append)))
 
 
 class TestBridgeHookImport(unittest.TestCase):
@@ -386,41 +454,47 @@ class TestBridgeHookImport(unittest.TestCase):
     @classmethod
     def setUpClass(cls):
         port = int(os.environ.get("TEST_PORT", bridge.DEFAULT_SERVER_PORT))
-        TestBridge.test_bridge = bridge.BridgeClient(
-            connect_to_port=port, loglevel=logging.DEBUG, hook_import=True)
+        cls.test_bridge = bridge.BridgeClient(
+            connect_to_port=port, loglevel=logging.DEBUG, hook_import=True, record_stats=True)
 
+    @print_stats
     def test_hook_import_top_level(self):
         """ Test that we handle import x syntax """
         import test_hook_import_top_level
         remote_name = str(test_hook_import_top_level)
         self.assertTrue("BridgedModule" in remote_name and "test_hook_import_top_level" in remote_name)
 
+    @print_stats
     def test_hook_import_dotted(self):
         """ Test that we handle import x.y syntax """
         import test_hook_import_dotted.child
         remote_name = str(test_hook_import_dotted.child)
         self.assertTrue("BridgedModule" in remote_name and "test_hook_import_dotted.child" in remote_name)
 
+    @print_stats
     def test_hook_import_from_syntax(self):
         """ Test that we handle from x import y syntax """
         from test_hook_import_from import run_server
         remote_name = str(run_server)
         self.assertTrue("BridgedCallable" in remote_name and "run_server" in remote_name)
 
+    @print_stats
     def test_hook_import_nonexistent(self):
         """ Test that we handle a nonexistent import """
         with self.assertRaises(ImportError):
             import foobar
 
+    @print_stats
     def test_hook_import_as(self):
         """ Test that we don't break import x as y syntax """
         import test_hook_import_as as thia
         remote_name = str(thia)
         self.assertTrue("BridgedModule" in remote_name and "test_hook_import_as" in remote_name)
 
+    @print_stats
     def test_hook_import_force_import(self):
         """ Test that we actually import something that's not loaded"""
-        remote_sys = TestBridge.test_bridge.remote_import("sys")
+        remote_sys = self.test_bridge.remote_import("sys")
         remote_python_version = remote_sys.version_info[0]
         local_python_version = sys.version_info[0]
 
@@ -444,6 +518,7 @@ class TestBridgeHookImport(unittest.TestCase):
             # same versions, can't think of anything useful to test load
             self.skipTest("Test irrelevant for matched versions")
 
+    @print_stats
     def test_local_import(self):
         """ Make sure a local import is resolved locally, not pulled in remotely """
         self.assertTrue("ast" not in sys.modules)
@@ -451,6 +526,7 @@ class TestBridgeHookImport(unittest.TestCase):
         name = str(ast)
         self.assertTrue("BridgedModule" not in name and "ast" in name)
         
+    @print_stats
     def test_hook_import_nonmodule(self):
         """ Test we can import nonmodules like modules (e.g., java classes from jython). But mostly so we can test 
             reimporting
@@ -472,8 +548,8 @@ class TestBridgeHookImportReimport(unittest.TestCase):
     def setUpClass(cls):
         port = int(os.environ.get("TEST_PORT", bridge.DEFAULT_SERVER_PORT))
         old_importer_index = len(sys.path)-1
-        TestBridge.test_bridge = bridge.BridgeClient(
-            connect_to_port=port, loglevel=logging.DEBUG, hook_import=True)
+        cls.test_bridge = bridge.BridgeClient(
+            connect_to_port=port, loglevel=logging.DEBUG, hook_import=True, record_stats=True)
 
         # rearrange paths to make sure our importer gets called first
         # TODO once we get around to implementing cleaning up import hooks on client shutdown, this shouldn't be required
@@ -482,6 +558,7 @@ class TestBridgeHookImportReimport(unittest.TestCase):
         sys.path[old_importer_index] = new_importer
         sys.path[-1] = old_importer
 
+    @print_stats
     def test_hook_import_nonmodule_again(self):
         """ If this fails with old/unknown handle, __spec__ has been set by the old client """
         # clear out our old import
@@ -497,12 +574,13 @@ class TestBridgeZZZZZZZShutdown(unittest.TestCase):
     @classmethod
     def setUpClass(cls):
         port = int(os.environ.get("TEST_PORT", bridge.DEFAULT_SERVER_PORT))
-        TestBridge.test_bridge = bridge.BridgeClient(
-            connect_to_port=port, loglevel=logging.DEBUG)
+        cls.test_bridge = bridge.BridgeClient(
+            connect_to_port=port, loglevel=logging.DEBUG, record_stats=True)
 
+    @print_stats
     def test_zzzzzz_shutdown(self):
         # test shutdown last
-        result = TestBridge.test_bridge.remote_shutdown()
+        result = self.test_bridge.remote_shutdown()
         self.assertTrue(result[bridge.SHUTDOWN])
 
         # give it a second to tear down
